"""In this example we plot a record's annotations on top of the curve of the
local GC content in the record's sequence.
"""
import matplotlib.pyplot as plt
from dna_features_viewer import BiopythonTranslator
from Bio import SeqIO
import numpy as np


def plot_local_gc_content(record, window_size, ax):
    """Plot windowed GC content on a designated Matplotlib ax."""
    def gc_content(s):
        return 100.0 * len([c for c in s if c in "GC"]) / len(s)

    yy = [
        gc_content(record.seq[i : i + window_size])
        for i in range(len(record.seq) - window_size)
    ]
    xx = np.arange(len(record.seq) - window_size) + 25
    ax.fill_between(xx, yy, alpha=0.3)
    ax.set_ylim(bottom=0)
    ax.set_ylabel("GC(%)")


record = SeqIO.read("example_sequence.gb", "genbank")
translator = BiopythonTranslator()
graphic_record = translator.translate_record(record)

fig, (ax1, ax2) = plt.subplots(
    2, 1, figsize=(10, 3), sharex=True, gridspec_kw={"height_ratios": [4, 1]}
)
graphic_record.plot(ax=ax1, with_ruler=False, strand_in_label_threshold=4)
plot_local_gc_content(record, window_size=50, ax=ax2)

fig.tight_layout()  # Resize the figure to the right height
fig.savefig("with_gc_plot.png")



import urllib
from io import StringIO
from dna_features_viewer import BiopythonTranslator, load_record

# DEFINE FEATURES ASPECTS


def features_properties(f):
    """Mutations get a red label, other features get a pastel color."""
    label = None
    if f.type == "Mutagenesis":
        label = f.qualifiers["Note"][0]
    color = {
        "Mutagenesis": "firebrick",
        "Active site": "yellow",
        "Beta strand": "lightyellow",
        "Chain": "lightcyan",
        "Helix": "honeydew",
        "Initiator methionine": "white",
        "Metal binding": "lightsteelblue",
        "Turn": "moccasin",
    }.get(f.type, "white")
    return dict(color=color, label=label)


# GET THE RECORD FROM UNIPROT

response = urllib.request.urlopen("https://www.uniprot.org/uniprot/P0A7B8.gff")
record_file = StringIO(response.read().decode())

# TRANSLATE AND PLOT THE RECORD

translator = BiopythonTranslator(features_properties=features_properties)
graphic_record = translator.translate_record(record_file)
ax, _ = graphic_record.plot(
    figure_width=15, max_label_length=100, elevate_outline_annotations=True,
)
ax.set_title("Mutation effects in P0A7B8", fontweight="bold", fontsize=16)
ax.figure.savefig("gff_record_from_the_web.png", bbox_inches="tight")



"""An example with GIF generation at the end. How cool is that!

This example requires the Moviepy library  installed (pip install moviepy).

"""
from Bio import Entrez, SeqIO
import moviepy.editor as mpe
from moviepy.video.io.bindings import mplfig_to_npimage
import matplotlib.pyplot as plt
from dna_features_viewer import BiopythonTranslator, CircularGraphicRecord

# DOWNLOAD THE PLASMID's RECORD FROM NCBI

handle = Entrez.efetch(
    db="nucleotide", id=1473096477, rettype="gb", retmode="text"
)
record = SeqIO.read(handle, "genbank")

# CREATE THE GRAPHIC RECORD WITH DNA_FEATURES_VIEWER

color_map = {
    "rep_origin": "yellow",
    "CDS": "orange",
    "regulatory": "red",
    "misc_recomb": "darkblue",
    "misc_feature": "lightblue",
}
translator = BiopythonTranslator(
    features_filters=(lambda f: f.type not in ["gene", "source"],),
    features_properties=lambda f: {"color": color_map.get(f.type, "white")},
)
translator.max_line_length = 15
graphic_record = translator.translate_record(
    record, record_class=CircularGraphicRecord
)
graphic_record.labels_spacing = 15

# ANIMATE INTO A GIF WITH MOVIEPY

duration = 5


def make_frame(t):
    top_nucleotide_index = t * graphic_record.sequence_length / duration
    graphic_record.top_position = top_nucleotide_index
    ax, _ = graphic_record.plot(figure_width=8, figure_height=11)
    ax.set_ylim(top=2)
    np_image = mplfig_to_npimage(ax.figure)
    plt.close(ax.figure)
    return np_image


clip = mpe.VideoClip(make_frame, duration=duration)
small_clip = clip.crop(x1=60, x2=-60, y1=100, y2=-100).resize(0.5)
small_clip.write_gif("example_with_gif.gif", fps=15)



""" dna_features_viewer/__init__.py """

from .GraphicRecord import GraphicRecord
from .CircularGraphicRecord import CircularGraphicRecord
from .GraphicFeature import GraphicFeature
from .BiopythonTranslator import (
    BiopythonTranslator,
    BlackBoxlessLabelTranslator,
)
from .biotools import load_record, annotate_biopython_record

from .version import __version__

__all__ = [
    "GraphicRecord",
    "CircularGraphicRecord",
    "GraphicFeature",
    "BiopythonTranslator",
    "BlackBoxlessLabelTranslator",
    "annotate_biopython_record",
    "__version__",
]



from .BiopythonTranslatorBase import BiopythonTranslatorBase


class BiopythonTranslator(BiopythonTranslatorBase):
    """A translator from SeqRecords to dna_features_viewer GraphicRecord.

    This can be subclassed to create custom "themes" (see the example
    ``custom_biopython_translator.py`` in the docs).

    This class is meant to be customized by subclassing and changing the
    methods (``compute_feature_label``, etc.) and/or the attributes
    (``default_feature_color`` etc).

    Attributes
    ----------

    default_feature_color = "#7245dc"
    graphic_record_parameters
      Dictionnary containing keyword arguments that will be passed to the
      (Circular)GraphicRecord constructor

    ignored_features_types
      A list or tuple of strings indicating all the feature types that should
      always be ignored (i.e. not included in the graphic record) by the
      translator

    label_fields
      This list of strings provides the order in which the different
      attributes of a Genbank feature will be considered, when automatically
      determining the feature label. For instance if the list is
      ["label", "source", "locus_tag"] and a feature has no label but has a
      "source", the "source" will be displayed in the plots.

    Parameters
    ----------

    features_filters
      List of filters (some_biopython_feature) => True/False.
      Only features passing all the filters are kept.
      This only works if you haven't redefined ``compute_filtered_features``

    features_properties
      A function (feature)=> properties_dict

    """

    default_feature_color = "#7245dc"
    ignored_features_types = ()
    label_fields = [
        "label",
        "name",
        "gene",
        "product",
        "source",
        "locus_tag",
        "note",
    ]

    def __init__(self, features_filters=(), features_properties=None):
        self.features_filters = features_filters
        self.features_properties = features_properties

    def compute_feature_color(self, feature):
        """Compute a color for this feature.

        If the feature has a ``color`` qualifier it will be used. Otherwise,
        the classe's ``default_feature_color`` is used.

        To change the behaviour, create a subclass of ``BiopythonTranslator``
        and overwrite this method.
        """
        if "color" in feature.qualifiers:
            color = feature.qualifiers["color"]
            if isinstance(color[0], str):
                return "".join(feature.qualifiers["color"])
            else:
                return color
        else:
            return self.default_feature_color

    def compute_feature_fontdict(self, feature):
        """Compute a font dict for this feature."""
        return None

    def compute_feature_box_linewidth(self, feature):
        """Compute a box_linewidth for this feature."""
        return 0.3

    def compute_feature_box_color(self, feature):
        """Compute a box_color for this feature."""
        return "auto"

    def compute_feature_label_link_color(self, feature):
        """Compute the color of the line linking the label to its feature."""
        return "black"

    def compute_filtered_features(self, features):
        """Return the list of features minus the ignored ones.
        
        By the method keeps any feature whose type is not in
        ignored_features_types and for which all filter(f) pass
        """
        return [
            f
            for f in features
            if all([fl(f) for fl in self.features_filters])
            and f.type not in self.ignored_features_types
        ]

    def compute_feature_label(self, feature):
        """Compute the label of the feature."""
        label = feature.type
        for key in self.label_fields:
            if key in feature.qualifiers and len(feature.qualifiers[key]):
                label = feature.qualifiers[key]
                break
        if isinstance(label, list):
            label = "|".join(label)
        return label

    def compute_feature_linewidth(self, feature):
        """Compute the edge width of the feature's arrow/rectangle."""
        return 1.0

    def compute_feature_legend_text(self, feature):
        return None

    def compute_feature_html(self, feature):
        """Gets the 'label' of the feature."""
        return self.compute_feature_label(feature)



"""Simple example where a few features are defined "by hand" and are displayed
and exported as PNG, first with a linear view, then with a circular
view.
"""

from dna_features_viewer import (
    GraphicFeature,
    GraphicRecord,
    CircularGraphicRecord,
)

features = [
    GraphicFeature(
        start=5, end=20, strand=+1, color="#ffd700", label="Small feature"
    ),
    GraphicFeature(
        start=20,
        end=500,
        strand=+1,
        color="#ffcccc",
        label="Gene 1 with a very long name",
    ),
    GraphicFeature(
        start=400, end=700, strand=-1, color="#cffccc", label="Gene 2"
    ),
    GraphicFeature(
        start=600, end=900, strand=+1, color="#ccccff", label="Gene 3"
    ),
]


# PLOT AND EXPORT A LINEAR VIEW OF THE CONSTRUCT
record = GraphicRecord(sequence_length=1000, features=features)
ax, _ = record.plot(figure_width=5)
ax.figure.savefig("graphic_record_defined_by_hand.png")


# PLOT AND EXPORT A CIRCULAR VIEW OF THE CONSTRUCT
circular_rec = CircularGraphicRecord(sequence_length=1000, features=features)
ax2, _ = circular_rec.plot(figure_width=4)
ax2.figure.tight_layout()
ax2.figure.savefig(
    "graphic_record_defined_by_hand_circular.png", bbox_inches="tight"
)



from ..biotools import find_narrowest_text_wrap

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import FeatureLocation, SeqFeature
from Bio.Alphabet import DNAAlphabet

from .MatplotlibPlottableMixin import MatplotlibPlottableMixin
from .BokehPlottableMixin import BokehPlottableMixin


class GraphicRecord(MatplotlibPlottableMixin, BokehPlottableMixin):
    """Set of Genetic Features of a same DNA sequence, to be plotted together.

    Parameters
    ----------

    sequence_length
      Length of the DNA sequence, in number of nucleotides

    features
      list of GraphicalFeature objects.

    feature_level_height
      Width in inches of one "level" for feature arrows.

    annotation_height
      Width in inches of one "level" for feature annotations.

    first_index
      Indicates the first index to plot in case the sequence is actually a
      subsequence of a larger one. For instance, if the Graphic record
      represents the segment (400, 420) of a sequence, we will have
      ``first_index=400`` and ``sequence_length=20``.

    plots_indexing
      Indicates which standard to use to show nucleotide indices in the plots.
      If 'biopython', the standard python indexing is used (starting at 0).
      If 'genbank', the indexing follows the Genbank standard (starting at 1).
    
    labels_spacing
      Number of pixels that will "pad" every labels to force some horizontal
      space between two labels or between a label and the borders of a feature.
    
    ticks_resolution
      Leave to "auto" for an auto-selected number of ticks on the ruler, or set
      to e.g. 50 for a tick every 50 nucleotide. 

    Attributes
    ----------

      default_font_family
        Default font to use for a feature that doesn't declare a font.
      
      default_ruler_color
        Default ruler color to use when no color is given at plot() time.
    
      default_box_color
        Default box color for non-inline annotations. If set to None, no
        boxes will be drawn unless the features declare a box_color.
        If "auto", a color (clearer version of the feature's color) will be
        computed, for all features also declaring their box_color as "auto".

      default_elevate_outline_annotations
        Value to use for elevate_outline_annotations when no specific value is
        given at ``graphic_record.plot(...)`` time. Set to true to have all
        text annotations appears above all features, or false else.
    """

    default_font_family = None
    default_ruler_color = "grey"
    default_box_color = "auto"
    min_y_height_of_text_line = 0.5

    def __init__(
        self,
        sequence_length=None,
        sequence=None,
        features=(),
        feature_level_height=1,
        first_index=0,
        plots_indexing="biopython",
        labels_spacing=8,
        ticks_resolution='auto'
    ):
        if sequence_length is None:
            sequence_length = len(sequence)
        self.features = features
        self.sequence_length = sequence_length
        self.feature_level_height = feature_level_height
        self.sequence = sequence
        self.first_index = first_index
        self.plots_indexing = plots_indexing
        self.labels_spacing = labels_spacing
        self.ticks_resolution = ticks_resolution

    @property
    def last_index(self):
        return self.first_index + self.sequence_length

    @property
    def span(self):
        """Return the display span (start, end) accounting for first_index."""
        return self.first_index, self.last_index

    def to_biopython_record(self, sequence):
        """
        Example
        -------
        from Bio import SeqIO
        gr_record = GraphicRecord(features=features, sequence_length=len(seq),
                                  sequence=seq)
        bio_record = gr_record.to_biopython_record()
        with open("example.gb", "w+") as f:
            SeqIO.write(record, f, "genbank")
        """
        features = [
            SeqFeature(
                FeatureLocation(f.start, f.end, f.strand),
                type=f.feature_type,
                qualifiers={"label": f.label},
            )
            for f in self.features
        ]
        if not isinstance(sequence, Seq):
            sequence = Seq(sequence, alphabet=DNAAlphabet())
        return SeqRecord(seq=sequence, features=features)

    def crop(self, window):
        start, end = window
        first_index = self.first_index
        if (start < first_index) or (end > self.last_index):
            raise ValueError("out-of-bound cropping")
        new_features = []
        for f in self.features:
            cropped_feature = f.crop(window)
            if cropped_feature is not None:  # = has ovelap with the window
                new_features.append(cropped_feature)

        return GraphicRecord(
            sequence=self.sequence[start - first_index : end - first_index]
            if self.sequence is not None
            else None,
            sequence_length=end - start,
            features=new_features,
            feature_level_height=self.feature_level_height,
            first_index=start,
            plots_indexing=self.plots_indexing,
            labels_spacing=self.labels_spacing,
            ticks_resolution=self.ticks_resolution
        )

    def determine_annotation_height(self, levels):
        """By default the ideal annotation level height is the same as the
        feature_level_height."""
        # TODO: Improve me! ideally, annotation width would be linked to the
        # height of one line of text, so dependent on font size and ax
        # height/span.
        return self.feature_level_height

    def coordinates_in_plot(self, x, level):
        """Convert a sequence position and height level into a (x, y) position.
        """
        return (x, level * self.feature_level_height)

    def split_overflowing_features_circularly(self):
        """Split the features that overflow over the edge for circular
        constructs (inplace)."""
        new_features = []
        for f in self.features:
            if f.start < 0 < f.end:
                f1, f2 = f.split_in_two(-1)
                f1.start, f1.end = (
                    f1.start + self.sequence_length,
                    f1.end + self.sequence_length,
                )
                new_features += [f1, f2]
            elif f.start < self.sequence_length < f.end:
                f1, f2 = f.split_in_two(self.sequence_length - 1)
                f2.start, f2.end = (
                    f2.start - self.sequence_length,
                    f2.end - self.sequence_length,
                )
                new_features += [f1, f2]
            else:
                new_features.append(f)
        self.features = new_features

    def _format_label(self, label, max_label_length=50, max_line_length=40):
        if len(label) > max_label_length:
            label = label[: max_label_length - 1] + "…"
        if len(label) > max_line_length:
            label = find_narrowest_text_wrap(label, max_line_length)
        return label

    def compute_padding(self, ax):
        ax_width = ax.get_window_extent(ax.figure.canvas.get_renderer()).width
        xmin, xmax = ax.get_xlim()
        return self.labels_spacing * (xmax - xmin) / (1.0 * ax_width)
